import unittest
import binascii
import hashlib
from neo3crypto import mmh3_hash_bytes, mmh3_hash
from neo3.core import types
from neo3.core import cryptography as crypto


class MerkleTreeTestCase(unittest.TestCase):
    def test_compute_root_single_hash(self):
        data = binascii.unhexlify(b"aa" * 32)
        hash1 = types.UInt256(data=data)
        root = crypto.MerkleTree.compute_root([hash1])

        self.assertEqual(data, root.to_array())

    def test_compute_root_multiple_hashes(self):
        expected_hash = hashlib.sha256(
            hashlib.sha256(binascii.unhexlify(b"aa" * 32 + b"bb" * 32)).digest()
        ).digest()

        hash1 = types.UInt256(data=binascii.unhexlify(b"aa" * 32))
        hash2 = types.UInt256(data=binascii.unhexlify(b"bb" * 32))
        hashes = [hash1, hash2]
        root = crypto.MerkleTree.compute_root(hashes)

        self.assertEqual(expected_hash, root.to_array())

    def test_computer_root_no_input(self):
        self.assertEqual(types.UInt256.zero(), crypto.MerkleTree.compute_root([]))

    def test_build_no_leaves(self):
        with self.assertRaises(ValueError) as context:
            crypto.MerkleTree._build([])
        self.assertIn("Leaves must have length", str(context.exception))

    def test_to_hash_array(self):
        hash1 = types.UInt256(data=binascii.unhexlify(b"aa" * 32))
        hash2 = types.UInt256(data=binascii.unhexlify(b"bb" * 32))
        hash3 = types.UInt256(data=binascii.unhexlify(b"cc" * 32))
        hash4 = types.UInt256(data=binascii.unhexlify(b"dd" * 32))
        hash5 = types.UInt256(data=binascii.unhexlify(b"ee" * 32))
        hashes = [hash1, hash2, hash3, hash4, hash5]

        m = crypto.MerkleTree(hashes)
        hash_array = m.to_hash_array()

        # sort the array
        hash_array = sorted(hash_array)

        for i, h in enumerate(hashes):
            self.assertEqual(h, hash_array[i])

    def test_merkle_node_methods(self):
        hash1 = types.UInt256(data=binascii.unhexlify(b"aa" * 32))
        hash2 = types.UInt256(data=binascii.unhexlify(b"bb" * 32))
        hashes = [hash1, hash2]
        m = crypto.MerkleTree(hashes)

        self.assertEqual(True, m.root.is_root())
        self.assertEqual(False, m.root.is_leaf())
        self.assertEqual(False, m.root.left_child.is_root())
        self.assertEqual(True, m.root.left_child.is_leaf())


class BloomFilterTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_seed_building(self):
        filter = crypto.BloomFilter(m=7, k=10, ntweak=123456)
        # these values have been captured by creating a BloomFilter in C# and reading the seed values through debugging
        expected_seeds = [
            123456,
            4222003669,
            4148916586,
            4075829503,
            4002742420,
            3929655337,
            3856568254,
            3783481171,
            3710394088,
            3637307005,
        ]
        self.assertEqual(expected_seeds, filter.seeds)

    def test_add_and_check(self):
        # modelled after https://github.com/neo-project/neo/blob/982e69090f27c1415872536ce39aea22f0873467/neo.UnitTests/Cryptography/UT_BloomFilter.cs#L12
        elements = b"\x00\x01\x02\x03\x04"
        filter = crypto.BloomFilter(m=7, k=10, ntweak=123456)
        filter.add(elements)
        self.assertTrue(filter.check(elements))
        self.assertFalse(filter.check(elements + b"\x05\x06\x07\x08\x09"))

    def test_init_with_elements(self):
        elements = b"\x00\x01\x02\x03\x04"
        m = 7
        k = 10
        filter = crypto.BloomFilter(m=m, k=k, ntweak=123456, elements=elements)
        self.assertEqual(m, len(filter.bits))
        self.assertEqual(k, len(filter.seeds))

    def test_get_bits(self):
        """
        byte[] elements = { 0, 1, 2, 3, 4};
        BloomFilter bf = new BloomFilter(7, 10, 123456, elements);
        byte[] buffer = new byte[(bf.M/8)+1];
        Console.WriteLine($"\\x{BitConverter.ToString(buffer).Replace("-", "\\x")}");
        """
        elements = b"\x00\x01\x02\x03\x04"
        filter = crypto.BloomFilter(m=7, k=10, ntweak=123456, elements=elements)
        self.assertEqual(b"\x00", filter.get_bits())


class Murmur128test(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_neo_cases(self):
        # https://github.com/Liaojinghui/neo/blob/30f33d075502acd792f804ffcf84cce689255306/tests/neo.UnitTests/Cryptography/UT_Murmur128.cs
        self.assertEqual(
            bytes.fromhex("0bc59d0ad25fde2982ed65af61227a0e"),
            mmh3_hash_bytes("hello", 123),
        )
        self.assertEqual(
            bytes.fromhex("3d3810fed480472bd214a14023bb407f"),
            mmh3_hash_bytes("world", 123),
        )
        self.assertEqual(
            bytes.fromhex("e0a0632d4f51302c55e3b3e48d28795d"),
            mmh3_hash_bytes("hello world", 123),
        )


class MurMur32test(unittest.TestCase):
    def test_one(self):
        x: int = mmh3_hash(b"\x01\x02\x03\x04\x05\x06", 123, signed=False)
        self.assertEqual(
            bytes.fromhex("3cdc1e41"), x.to_bytes(4, "little", signed=False)
        )



"""
Customizable payload.
"""
from __future__ import annotations
import hashlib
from neo3.core import types, serialization, Size as s, utils
from neo3.network.payloads import inventory, verification


class ExtensiblePayload(inventory.IInventory):
    """
    A payload that supports arbitrary `data`.
    """

    # Comes from neo3.network.message.Message.PAYLOAD_MAX_SIZE but wanted to break the import cycle
    # despite not causing any circular imported errors yet
    PAYLOAD_MAX_SIZE = 0x2000000

    def __init__(
        self,
        category: str,
        valid_block_start: int,
        valid_block_end: int,
        sender: types.UInt160,
        data: bytes,
        witness: verification.Witness,
    ):
        super(ExtensiblePayload, self).__init__()
        #: An identifier to which category the data belongs
        self.category = category
        #: Starting height in which the payload is valid
        self.valid_block_start = valid_block_start
        #: Last height in which the payload is valid
        self.valid_block_end = valid_block_end
        #: The hash of the account who has sent the payload to the network
        self.sender = sender
        #: Arbitrary data as required by the payload category
        self.data = data
        #: The witness of the payload
        self.witness = witness

        self.witnesses = [self.witness]  # for IVerifiable super

    def __len__(self):
        return (
            utils.get_var_size(self.category)
            + s.uint32
            + s.uint32
            + s.uint160
            + utils.get_var_size(self.data)
            + 1
            + len(self.witness)
        )

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_uint8(1)
        writer.write_serializable(self.witness)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the unsigned part of the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_string(self.category)
        writer.write_uint32(self.valid_block_start)
        writer.write_uint32(self.valid_block_end)
        writer.write_serializable(self.sender)
        writer.write_var_bytes(self.data)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the check byte is not 1.
        """
        self.deserialize_unsigned(reader)
        if reader.read_uint8() != 1:
            raise ValueError("Deserialization error - check byte incorrect")
        self.witness = reader.read_serializable(verification.Witness)

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the unsigned data part of the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the valid_block_start exceeds the valid_block_end field.
        """
        self.category = reader.read_var_string(32)
        self.valid_block_start = reader.read_uint32()
        self.valid_block_end = reader.read_uint32()
        if self.valid_block_start >= self.valid_block_end:
            raise ValueError(
                "Deserialization error - valid_block_start is bigger than valid_block_end"
            )
        self.sender = reader.read_serializable(types.UInt160)
        self.data = reader.read_var_bytes(self.PAYLOAD_MAX_SIZE)

    def hash(self) -> types.UInt256:
        """
        Get a unique identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(data_to_hash).digest()
            return types.UInt256(data=data)

    @property
    def inventory_type(self) -> inventory.InventoryType:
        """
        Inventory type identifier.
        """
        return inventory.InventoryType.EXTENSIBLE

    def get_script_hashes_for_verifying(self, snapshot) -> list[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        return [self.sender]

    @classmethod
    def _serializable_init(cls):
        return cls("", 0, 0, types.UInt160.zero(), b"", verification.Witness(b"", b""))



import json
import os.path
import unittest

from neo3.wallet import wallet, scrypt_parameters as scrypt, account


class WalletCreationTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_wallet_new_wallet(self):
        wallet_file_name = "unittest-wallet"
        wallet_file_path = f"{wallet_file_name}.json"

        # remove the file if it exists for proper testing
        if os.path.isfile(wallet_file_path):
            os.remove(wallet_file_path)

        test_wallet = wallet.DiskWallet(wallet_file_name)
        scrypt_parameters_default = scrypt.ScryptParameters()

        self.assertEqual(wallet_file_name, test_wallet.name)
        self.assertEqual("1.0", test_wallet.version)
        self.assertEqual(scrypt_parameters_default.n, test_wallet.scrypt.n)
        self.assertEqual(scrypt_parameters_default.r, test_wallet.scrypt.r)
        self.assertEqual(scrypt_parameters_default.p, test_wallet.scrypt.p)
        self.assertEqual([], test_wallet.accounts)
        self.assertEqual({}, test_wallet.extra)

    def test_wallet_default_value(self):
        test_wallet = wallet.DiskWallet.default()
        scrypt_parameters_default = scrypt.ScryptParameters()

        self.assertEqual("wallet.json", test_wallet.name)
        self.assertEqual("1.0", test_wallet.version)
        self.assertEqual(scrypt_parameters_default.n, test_wallet.scrypt.n)
        self.assertEqual(scrypt_parameters_default.r, test_wallet.scrypt.r)
        self.assertEqual(scrypt_parameters_default.p, test_wallet.scrypt.p)
        self.assertEqual([], test_wallet.accounts)
        self.assertEqual({}, test_wallet.extra)

    def test_wallet_save(self):
        wallet_path = "unittest-wallet-save.json"
        # remove the file if it exists for proper testing
        if os.path.isfile(wallet_path):
            os.remove(wallet_path)

        test_wallet = wallet.DiskWallet.default(wallet_path, "NEP6 Wallet")
        test_wallet.save()
        self.assertTrue(os.path.isfile(wallet_path))

        with open(wallet_path) as json_file:
            data = json.load(json_file)

        self.assertEqual(data["name"], test_wallet.name)
        self.assertEqual(data["version"], test_wallet.version)
        self.assertEqual(data["scrypt"]["n"], test_wallet.scrypt.n)
        self.assertEqual(data["scrypt"]["r"], test_wallet.scrypt.r)
        self.assertEqual(data["scrypt"]["p"], test_wallet.scrypt.p)
        self.assertEqual(data["accounts"], test_wallet.accounts)
        self.assertEqual({}, test_wallet.extra)
        self.assertEqual(None, data["extra"])

        # remove the file if it exists for proper testing
        if os.path.isfile(wallet_path):
            os.remove(wallet_path)

        # save using context manager
        with wallet.DiskWallet.default(wallet_path, "NEP6 Wallet"):
            pass
        self.assertTrue(os.path.isfile(wallet_path))

        # clean up after test
        if os.path.isfile(wallet_path):
            os.remove(wallet_path)

    def test_wallet_from_json(self):
        password = "123"

        new_wallet = wallet.DiskWallet.default()
        # override scrypt parameters for testing
        new_wallet.scrypt = scrypt.ScryptParameters(2, 8, 8)
        test_account = account.Account.create_new(
            password, scrypt_parameters=scrypt.ScryptParameters(2, 8, 8)
        )
        new_wallet.account_add(test_account, is_default=True)

        json_wallet = new_wallet.to_json()

        test_wallet = wallet.Wallet.from_json(json_wallet, passwords=["123"])
        self.assertEqual(new_wallet.name, test_wallet.name)
        self.assertEqual("1.0", test_wallet.version)
        self.assertEqual(1, len(test_wallet.accounts))
        self.assertEqual(test_account, test_wallet.accounts[0])
        self.assertEqual(test_wallet._default_account, test_wallet.accounts[0])

    def test_wallet_account_new(self):
        password = "abcabc"
        test_wallet = wallet.DiskWallet.default()
        # override scrypt parameters for testing
        test_wallet.scrypt = scrypt.ScryptParameters(2, 8, 8)
        self.assertEqual(0, len(test_wallet.accounts))

        # create account without label
        acc = test_wallet.account_new(password)
        self.assertEqual(1, len(test_wallet.accounts))
        self.assertEqual(None, acc.label)
        self.assertEqual(test_wallet._default_account, acc)

        # create account with label
        label = "New Account"
        acc = test_wallet.account_new(password, label)
        self.assertEqual(2, len(test_wallet.accounts))
        self.assertEqual(label, acc.label)
        self.assertNotEqual(test_wallet._default_account, acc)

        # create account with duplicated label
        with self.assertRaises(ValueError) as context:
            test_wallet.account_new(password, label)
        self.assertIn("Label is already used by an account", str(context.exception))

        # create account and set as default
        label = "Other Account"
        acc = test_wallet.account_new(password, label, is_default=True)
        self.assertEqual(3, len(test_wallet.accounts))
        self.assertEqual(label, acc.label)
        self.assertEqual(test_wallet._default_account, acc)

    def test_wallet_account_add(self):
        password = "abcabc"
        test_wallet = wallet.DiskWallet.default()

        self.assertEqual(0, len(test_wallet.accounts))

        scryptp = scrypt.ScryptParameters(2, 8, 8)
        label = "New Account"
        account_1 = account.Account(password=password, scrypt_parameters=scryptp)
        account_2 = account.Account(
            password=password, label=label, scrypt_parameters=scryptp
        )
        account_3 = account.Account(password=password)
        account_4 = account.Account(
            password=password, label=label, scrypt_parameters=scryptp
        )

        # add account, first account is set as default
        success = test_wallet.account_add(account_1)
        self.assertTrue(success)
        self.assertEqual(1, len(test_wallet.accounts))
        self.assertEqual(test_wallet._default_account, account_1)

        # add account
        success = test_wallet.account_add(account_2)
        self.assertTrue(success)
        self.assertEqual(2, len(test_wallet.accounts))
        self.assertNotEqual(test_wallet._default_account, account_2)

        # add account already added
        success = test_wallet.account_add(account_2)
        self.assertFalse(success)

        # add account and set it as default
        success = test_wallet.account_add(account_3, is_default=True)
        self.assertTrue(success)
        self.assertEqual(3, len(test_wallet.accounts))
        self.assertEqual(test_wallet._default_account, account_3)

        # add account with duplicated label
        with self.assertRaises(ValueError) as context:
            test_wallet.account_add(account_4)
        self.assertIn("Label is already used by an account", str(context.exception))

    def test_wallet_account_delete(self):
        scryptp = scrypt.ScryptParameters(2, 8, 8)
        password = "abcabc"
        account_1 = account.Account(password=password, scrypt_parameters=scryptp)
        account_2 = account.Account(password=password, scrypt_parameters=scryptp)
        account_3 = account.Account(password=password, scrypt_parameters=scryptp)

        test_wallet = wallet.DiskWallet.default()
        test_wallet.account_add(account_1)
        test_wallet.account_add(account_2)
        test_wallet.account_add(account_3)

        self.assertEqual(account_1, test_wallet._default_account)

        # delete account that is not default
        success = test_wallet.account_delete(account_2)
        self.assertTrue(success)
        self.assertEqual(account_1, test_wallet._default_account)

        # delete account not included
        success = test_wallet.account_delete(account_2)
        self.assertFalse(success)

        # delete account default, with other existing
        success = test_wallet.account_delete(account_1)
        self.assertTrue(success)
        self.assertNotEqual(account_1, test_wallet._default_account)
        self.assertEqual(account_3, test_wallet._default_account)

        # delete account default and it's the only existing account
        success = test_wallet.account_delete(account_3)
        self.assertTrue(success)
        self.assertIsNone(test_wallet._default_account)

    def test_wallet_account_delete_by_label(self):
        label_1 = "Account 1"
        label_2 = "Account 2"
        label_not_used = "Account 3"

        password = "123123"
        scryptp = scrypt.ScryptParameters(2, 8, 8)
        account_1 = account.Account(
            password=password, label=label_1, scrypt_parameters=scryptp
        )
        account_2 = account.Account(
            password=password, label=label_2, scrypt_parameters=scryptp
        )
        account_3 = account.Account(password=password, scrypt_parameters=scryptp)

        test_wallet = wallet.DiskWallet.default()
        test_wallet.account_add(account_1)
        test_wallet.account_add(account_2)
        test_wallet.account_add(account_3)

        self.assertEqual(account_1, test_wallet._default_account)

        # delete by label when account is not the default
        success = test_wallet.account_delete_by_label(label_2)
        self.assertTrue(success)
        self.assertEqual(account_1, test_wallet._default_account)

        # delete label not included
        success = test_wallet.account_delete_by_label(label_not_used)
        self.assertFalse(success)

        # delete by label when account is default
        success = test_wallet.account_delete_by_label(label_1)
        self.assertTrue(success)
        self.assertNotEqual(account_1, test_wallet._default_account)
        self.assertEqual(account_3, test_wallet._default_account)

    def test_from_json_with_multisig_account(self):
        p = os.path.join(os.path.dirname(__file__), "rc2-wallet.json")
        with open(p) as f:
            data = json.load(f)

        w = wallet.Wallet.from_json(data, ["123", "123"])
        self.assertEqual(2, len(w.accounts))
        self.assertEqual("NY9qiu8YScTM9oAc3nnaeNjaX5fnraaRTA", w.accounts[0].address)
        self.assertEqual("NcmoFiYqThZJFiEYVF1BjYEk6YwF5vtkFA", w.accounts[1].address)

    def test_from_json_with_multiple_accounts(self):
        label_1 = "Account 1"
        label_2 = "Account 2"

        password1 = "123123"
        password2 = "456456"
        scryptp = scrypt.ScryptParameters(2, 8, 8)
        account_1 = account.Account(
            password=password1, label=label_1, scrypt_parameters=scryptp
        )
        account_2 = account.Account(
            password=password2, label=label_2, scrypt_parameters=scryptp
        )
        w = wallet.Wallet(name="test wallet", scrypt_params=scryptp)
        w.account_add(account_1)
        w.account_add(account_2)
        w_json = w.to_json()

        # now test we can load it from_json
        w2_with_passwords = wallet.Wallet.from_json(w_json, [password1, password2])
        self.assertEqual(2, len(w2_with_passwords.accounts))
        self.assertFalse(w2_with_passwords.accounts[0].is_watchonly)
        self.assertFalse(w2_with_passwords.accounts[0].is_watchonly)

        w2_as_watchonly = wallet.Wallet.from_json(w_json)
        self.assertEqual(2, len(w2_with_passwords.accounts))
        self.assertTrue(w2_as_watchonly.accounts[0].is_watchonly)
        self.assertTrue(w2_as_watchonly.accounts[0].is_watchonly)

    def test_insufficient_passwords_provided_from_json(self):
        p = os.path.join(os.path.dirname(__file__), "rc2-wallet.json")
        with open(p) as f:
            data = json.load(f)

        with self.assertRaises(ValueError) as context:
            wallet.Wallet.from_json(data, ["123"])
        self.assertIn(
            "Incorrect number of passwords provided (1) for number of accounts in wallet (2)",
            str(context.exception),
        )



"""
Smart contract and account contract classes. Contains a list of all native contracts.
"""
from __future__ import annotations
from collections.abc import Sequence
from dataclasses import dataclass
from neo3.contracts import abi, utils, nef, manifest
from neo3.core import cryptography, utils as coreutils, types, serialization, Size as s


@dataclass
class _ContractHashes:
    CRYPTO_LIB = types.UInt160.from_string("0x726cb6e0cd8628a1350a611384688911ab75f51b")
    GAS_TOKEN = types.UInt160.from_string("0xd2a4cff31913016155e38e474a2c06d08be276cf")
    LEDGER = types.UInt160.from_string("0xda65b600f7124ce6c79950c1772a36403104f2be")
    MANAGEMENT = types.UInt160.from_string("0xfffdc93764dbaddd97c48f252a53ea4643faa3fd")
    NEO_TOKEN = types.UInt160.from_string("0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5")
    ORACLE = types.UInt160.from_string("0xfe924b7cfe89ddd271abaf7210a80a7e11178758")
    POLICY = types.UInt160.from_string("0xcc5e4edd9f5f8dba8bb65734541df7a1c081c67b")
    ROLE_MANAGEMENT = types.UInt160.from_string(
        "0x49cf4e5378ffcd4dec034fd98a174c5491e395e2"
    )
    STD_LIB = types.UInt160.from_string("0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0")


#: List of Neo's native contract hashes.
CONTRACT_HASHES = _ContractHashes()


class Contract:
    """
    Generic contract.
    """

    def __init__(
        self, script: bytes, parameter_list: Sequence[abi.ContractParameterType]
    ):
        #: The contract instructions (OpCodes)
        self.script = script
        self.parameter_list = parameter_list
        self._script_hash = coreutils.to_script_hash(self.script)
        self._address = None

    @property
    def script_hash(self) -> types.UInt160:
        """
        The contract script hash.
        """
        return self._script_hash

    @classmethod
    def create_multisig_contract(
        cls, m: int, public_keys: Sequence[cryptography.ECPoint]
    ) -> Contract:
        """
        Create a multi-signature contract requiring `m` signatures from the list `public_keys`.

        Args:
            m: minimum number of signature required for signing. Can't be lower than 2.
            public_keys: public keys to use during verification.
        """
        return cls(
            script=utils.create_multisig_redeemscript(m, public_keys),
            parameter_list=[abi.ContractParameterType.SIGNATURE] * m,
        )

    @classmethod
    def create_signature_contract(cls, public_key: cryptography.ECPoint) -> Contract:
        """
        Create a signature contract.

        Args:
            public_key: the public key to use during verification.
        """
        return cls(
            utils.create_signature_redeemscript(public_key),
            [abi.ContractParameterType.SIGNATURE],
        )


class ContractState(serialization.ISerializable):
    """
    Smart contract chain state container.
    """

    def __init__(
        self,
        id_: int,
        nef: nef.NEF,
        manifest_: manifest.ContractManifest,
        update_counter: int,
        hash_: types.UInt160,
    ):
        self.id = id_
        self.nef = nef
        self.manifest = manifest_
        self.update_counter = update_counter
        self.hash = hash_

    def __len__(self):
        return (
            s.uint32  # id
            + len(self.nef.to_array())
            + len(self.manifest)
            + s.uint16  # update counter
            + len(self.hash)
        )

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash != other.hash:
            return False
        return True

    def __deepcopy__(self, memodict={}):
        return ContractState.deserialize_from_bytes(self.to_array())

    @property
    def script(self) -> bytes:
        """
        NEF script
        """
        return self.nef.script

    @script.setter
    def script(self, value: bytes) -> None:
        self.nef.script = value

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_int32(self.id)
        writer.write_serializable(self.nef)
        writer.write_serializable(self.manifest)
        writer.write_uint16(self.update_counter)
        writer.write_serializable(self.hash)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.id = reader.read_int32()
        self.nef = reader.read_serializable(nef.NEF)
        self.manifest = reader.read_serializable(manifest.ContractManifest)
        self.update_counter = reader.read_uint16()
        self.hash = reader.read_serializable(types.UInt160)

    def can_call(self, target_contract: ContractState, target_method: str) -> bool:
        """
        Utility function to check if the contract has permission to call `target_method` on `target_contract`.

        Args:
            target_contract:
            target_method:

        Returns:
            `True` if allowed. `False` if not possible.
        """
        results = list(
            map(
                lambda p: p.is_allowed(
                    target_contract.hash, target_contract.manifest, target_method
                ),
                self.manifest.permissions,
            )
        )
        return any(results)

    @classmethod
    def _serializable_init(cls):
        return cls(
            0,
            nef.NEF._serializable_init(),
            manifest.ContractManifest(),
            0,
            types.UInt160.zero(),
        )



import unittest
from neo3.api.helpers import stdlib
from neo3.core import types


class TestStdLibHelpers(unittest.TestCase):
    def test_with_map(self):
        """
        C# reference code
            var m = new Map(new ReferenceCounter());
            var i = new Integer(new BigInteger(1));
            var i2 = new Integer(new BigInteger(2));
            var b = new Neo.VM.Types.Boolean(true);
            m[i] = b;
            m[i2] = b;
        """
        # moved outside of multiline comment because pycharm is broken: https://youtrack.jetbrains.com/issue/PY-43117
        # Console.WriteLine($"b'\\x{BitConverter.ToString(BinarySerializer.Serialize(m, 999)).Replace("-", @"\x")}'");

        data = b"\x48\x02\x21\x01\x01\x20\x01\x21\x01\x02\x20\x01"

        expected = {1: True, 2: True}
        results: dict = stdlib.binary_deserialize(data)

        self.assertEqual(expected, results)

    def test_with_array(self):
        """
            var a = new Neo.VM.Types.Array();
        var i = new Integer(new BigInteger(1));
        var i2 = new Integer(new BigInteger(2));
        a.Add(i);
        a.Add(i2);
        """
        # Console.WriteLine($"b'\\x{BitConverter.ToString(BinarySerializer.Serialize(a, 999)).Replace("-", @"\x")}'");
        # moved outside of multiline comment because pycharm is broken: https://youtrack.jetbrains.com/issue/PY-43117
        data = b"\x40\x02\x21\x01\x01\x21\x01\x02"
        expected = [1, 2]
        results: dict = stdlib.binary_deserialize(data)
        self.assertEqual(expected, results)

    def test_with_null(self):
        data = b"\x00"
        expected = None
        results: dict = stdlib.binary_deserialize(data)
        self.assertEqual(expected, results)

    def test_deserialize_bytestring(self):
        data = b"\x28\x02\x01\x02"
        expected = b"\x01\x02"
        results: dict = stdlib.binary_deserialize(data)
        self.assertEqual(expected, results)

    def test_deserialize_buffer(self):
        data = b"\x30\x02\x01\x02"
        expected = b"\x01\x02"
        results: dict = stdlib.binary_deserialize(data)
        self.assertEqual(expected, results)

    def test_deserialize_struct(self):
        # struct with 2 integers (1,2)
        data = b"\x41\x02\x21\x01\x01\x21\x01\x02"
        expected = [1, 2]
        results: dict = stdlib.binary_deserialize(data)
        self.assertEqual(expected, results)

    def test_invalid(self):
        data = b"\xFF"  # invalid stack item type
        with self.assertRaises(ValueError) as context:
            stdlib.binary_deserialize(data)
        self.assertIn("not a valid StackItemType", str(context.exception))

        with self.assertRaises(ValueError) as context:
            stdlib.binary_deserialize(b"")
        self.assertEqual("Nothing to deserialize", str(context.exception))


if __name__ == "__main__":
    unittest.main()



"""
NEO address utilities.
"""
import base58
from neo3.core import types, cryptography, utils as coreutils
from neo3.wallet.types import NeoAddress
from neo3.contracts import utils as contractutils


def script_hash_to_address(
    script_hash: types.UInt160, address_version: int = 0x35
) -> NeoAddress:
    """
    Convert the specified script hash to an address.

    Args:
        script_hash: script hash to convert.
        address_version: network protocol address version. Historically has been fixed to `0x35` for MainNet and TestNet.
         Use the `getversion()` RPC method to query for its value.
    """
    data = address_version.to_bytes(1, "little") + script_hash.to_array()
    return base58.b58encode_check(data).decode("utf-8")


def address_to_script_hash(address: NeoAddress) -> types.UInt160:
    """
    Convert the specified address to a script hash.

    Args:
        address: address to convert.

    Raises:
        ValueError: if the length of data (address value in bytes) is not valid.
        ValueError: if the account version is not valid.
    """
    validate_address(address)
    data = base58.b58decode_check(address)
    return types.UInt160(data[1:])


def public_key_to_script_hash(public_key: cryptography.ECPoint) -> types.UInt160:
    """
    Convert the specified public key to a script hash.
    """
    contract_script = contractutils.create_signature_redeemscript(public_key)
    return coreutils.to_script_hash(contract_script)


def is_valid_address(address: NeoAddress) -> bool:
    """
    Test if the provided address is a valid address.

    Args:
        address: an address.
    """
    try:
        validate_address(address)
    except ValueError:
        return False
    return True


def validate_address(address: NeoAddress, address_version: int = 0x35) -> None:
    """
    Validate a given address. If address is not valid an exception will be raised.

    Args:
        address: an address.
        address_version: network protocol address version. Historically has been fixed to `0x35` for MainNet and TestNet.
         Use the `getversion()` RPC method to query for its value.

    Raises:
        ValueError: if the length of data(address value in bytes) is not valid.
        ValueError: if the account version is not valid.
    """
    data: bytes = base58.b58decode_check(address)
    if len(data) != len(types.UInt160.zero()) + 1:
        raise ValueError(
            f"The address is wrong, because data (address value in bytes) length should be "
            f"{len(types.UInt160.zero()) + 1}"
        )
    elif data[0] != address_version:
        raise ValueError(f"The account version is not {address_version}")



"""
    This version of the contract has an `add` method that increases the value by 1.
"""
from typing import Any
from boa3.builtin import NeoMetadata, metadata, public
from boa3.builtin.nativecontract.contractmanagement import ContractManagement


@metadata
def manifest_metadata() -> NeoMetadata:
    """
    Defines this smart contract's metadata information.
    """
    meta = NeoMetadata()
    meta.name = "Example Contract"
    return meta


@public(safe=False)
def update(nef_file: bytes, manifest: bytes, data: Any = None):
    ContractManagement.update(nef_file, manifest, data)


@public
def add(number: int) -> int:
    return number + 1


@public(safe=False)
def destroy():
    ContractManagement.destroy()



import pathlib
import os
import sys
import subprocess
import threading
import shlex
import time
import json
from typing import Optional
from neo3.wallet.wallet import Wallet
from neo3.core import types

shared_dir = pathlib.Path("shared").resolve(strict=True)

user_wallet = Wallet.from_file(f"{shared_dir}/user-wallet.json", passwords=["123"])
coz_wallet = Wallet.from_file(f"{shared_dir}/coz-wallet.json", passwords=["123"])
neoxpress_config_path = f"{shared_dir}/default.neo-express"
neoxpress_batch_path = f"{shared_dir}/setup-neoxp-for-tests.batch"
coz_token_hash = types.UInt160.from_string("0x41ee5befd936c90f15893261abbd681f20ed0429")
# corresponds to the nep-11 token in the `/nep11-token/` dir and deployed with the `coz` account
nep11_token_hash = types.UInt160.from_string(
    "0x35de2913c480c19a7667da1cc3b2fe3e4c9de761"
)


class NeoExpress:
    """Neo express wrapper"""

    def __init__(
        self,
        config_path: Optional[str] = None,
        batch_path: Optional[str] = None,
        executable_path: Optional[str] = None,
        debug: bool = False,
    ):
        # The full qualified path is needed for correctly capturing streaming stdout with Popen without shell=True
        # The assumption is that the tool is installed in the standard location (as per
        # https://learn.microsoft.com/en-us/dotnet/core/tools/global-tools#install-a-global-tool )
        # Unless specified manually
        if sys.platform == "win32":
            self.prog = f"{pathlib.Path().home()}\\.dotnet\\tools\\neoxp"
        else:
            self.prog = f"{pathlib.Path().home()}/.dotnet/tools/neoxp"

        self.config_path = (
            config_path if config_path is not None else neoxpress_config_path
        )
        self.batch_path = batch_path if batch_path is not None else neoxpress_batch_path
        self.debug = debug

        self._process = None
        self._ready = False
        self._stop = False

        with open(self.config_path) as f:
            data = json.load(f)
            port = data["consensus-nodes"][0]["rpc-port"]
            address = data["settings"].get("rpc.BindAddress", "127.0.0.1")
            self.rpc_host = f"http://{address}:{port}"

        if executable_path is not None:
            self._verify_executable(executable_path)
            self.prog = executable_path
        else:
            try:
                subprocess.run(
                    [self.prog, "-h"],
                    check=True,
                    stdout=subprocess.DEVNULL,
                )
            except subprocess.SubprocessError:
                raise ValueError(
                    "Cannot automatically neo express executable. Please specify the path"
                )

    def _verify_executable(self, full_path: str):
        if not os.path.isfile(full_path) or not os.access(full_path, os.X_OK):
            raise ValueError(f"Invalid executable: {full_path}")

    def initialize_with(self, batch_path: str):
        print("executing neo-express batch...", end="")
        cmd = f"{self.prog} batch -r {batch_path}"
        subprocess.run(shlex.split(cmd), check=True, stdout=subprocess.DEVNULL)
        print("done")

    def run(self):
        print("starting neo-express...", end="")
        cmd = f"{self.prog} run -i {self.config_path}"

        self._process = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
            text=True,
            shell=False,
        )

        def process_stdout(process):
            for output in iter(process.stdout.readline, b""):
                if self._stop is True:
                    break

                if "Neo express is running" in output:
                    self._ready = True

                if self.debug:
                    print(output, end="")

        thread = threading.Thread(target=process_stdout, args=(self._process,))
        thread.start()

        while not self._ready:
            time.sleep(0.0001)

        print("done")

    def stop(self):
        print("stopping neo-express...", end="")
        # break out of the process_stdout loop
        self._stop = True
        self._process.kill()
        print("done")

    @classmethod
    def at(
        cls,
        executable_path: str,
        config_path: Optional[str] = None,
        batch_path: Optional[str] = None,
        debug: bool = False,
    ):
        return cls(config_path, batch_path, executable_path, debug)

    def __enter__(self):
        if self.batch_path is not None:
            self.initialize_with(self.batch_path)
        self.run()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



import json
import binascii
from types import SimpleNamespace
from neo3.core import cryptography


class IndexableNamespace(SimpleNamespace):
    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __contains__(self, key):
        try:
            self.__dict__[key]
            return True
        except KeyError:
            return False

    def get(self, key, default=None):
        try:
            return self.__dict__[key]
        except KeyError:
            return default


class Settings(IndexableNamespace):
    db = None
    _cached_standby_committee = None
    default_settings = {
        "network": {
            "magic": 5195086,
            "account_version": 53,
            "seedlist": [],
            "validators_count": 1,
            "standby_committee": [
                "02158c4a4810fa2a6a12f7d33d835680429e1a68ae61161c5b3fbc98c7f1f17765"
            ],
        },
    }

    @classmethod
    def from_json(cls, json: dict):
        o = cls(**json)
        o._convert(o.__dict__, o.__dict__)
        return o

    @classmethod
    def from_file(cls, path_to_json: str):
        with open(path_to_json, "r") as f:
            data = json.load(f)
        return cls.from_json(data)

    def register(self, json: dict):
        self.__dict__.update(json)
        self._convert(self.__dict__, self.__dict__)

    def _convert(self, what: dict, where: dict):
        # turn all _dictionary what into IndexableNamespaces
        to_update = []
        for k, v in what.items():
            if isinstance(v, dict):
                to_update.append((k, IndexableNamespace(**v)))

        for k, v in to_update:
            if isinstance(where, dict):
                where.update({k: v})
            else:
                where.__dict__.update({k: v})
            self._convert(where[k].__dict__, where[k].__dict__)

    @property
    def standby_committee(self) -> list[cryptography.ECPoint]:
        if self._cached_standby_committee is None:
            points = []
            for p in self.network.standby_committee:
                points.append(
                    cryptography.ECPoint.deserialize_from_bytes(binascii.unhexlify(p))
                )
            self._cached_standby_committee = points
        return self._cached_standby_committee

    @property
    def standby_validators(self) -> list[cryptography.ECPoint]:
        return self.standby_committee[: self.network.validators_count]

    def reset_settings_to_default(self):
        self.__dict__.clear()
        self.__dict__.update(self.from_json(self.default_settings).__dict__)


settings = Settings.from_json(Settings.default_settings)



import unittest
import binascii
import enum
from copy import deepcopy
from bitarray import bitarray
from neo3.network import capabilities
from neo3.network.payloads import (
    address,
    empty,
    transaction,
    block,
    verification,
    inventory,
    ping,
    filter as filterpayload,
    version,
)
from neo3.core import types, serialization, cryptography as crypto
from neo3.settings import settings


class AddrTestCase(unittest.TestCase):
    """
    Tests both Network addresses and the AddrPayload
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        var capability = new FullNodeCapability(123);

        var ip = IPAddress.Parse("127.0.0.1");
        var network_addr = NetworkAddressWithTime.Create(ip, 0, capability);
        Console.WriteLine($"len: {network_addr.Size}");
        Console.WriteLine($"b\'{BitConverter.ToString(network_addr.ToArray()).Replace("-", "")}\'");
        """
        cls.network_addr = address.NetworkAddress(
            address="127.0.0.1:0",
            timestamp=0,
            capabilities=[capabilities.FullNodeCapability(start_height=123)],
        )

        cls.addr_payload = address.AddrPayload([cls.network_addr])

    def test_network_addr_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 26
        self.assertEqual(expected_len, len(self.network_addr))

    def test_network_addr_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"0000000000000000000000000000FFFF7F00000101107B000000"
        )
        self.assertEqual(expected_data, self.network_addr.to_array())

    def test_network_addr_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_addr = address.NetworkAddress.deserialize_from_bytes(
            self.network_addr.to_array()
        )
        self.assertEqual(self.network_addr.timestamp, deserialized_addr.timestamp)
        self.assertEqual(self.network_addr.address, deserialized_addr.address)
        self.assertEqual(
            self.network_addr.capabilities[0].start_height,
            deserialized_addr.capabilities[0].start_height,
        )

    def test_addrpayload_len(self):
        expected_len = 27
        self.assertEqual(expected_len, len(self.addr_payload))

    def test_addrpayload_serialization(self):
        expected_data = binascii.unhexlify(
            b"010000000000000000000000000000FFFF7F00000101107B000000"
        )
        self.assertEqual(expected_data, self.addr_payload.to_array())

    def test_addrpayload_deserialization(self):
        deserialized_addr_payload = address.AddrPayload.deserialize_from_bytes(
            self.addr_payload.to_array()
        )
        self.assertEqual(
            len(self.addr_payload.addresses), len(deserialized_addr_payload.addresses)
        )
        self.assertEqual(
            self.addr_payload.addresses[0].capabilities[0].start_height,
            deserialized_addr_payload.addresses[0].capabilities[0].start_height,
        )

    def test_addrpayload_deserialization2(self):
        capa = capabilities.ServerCapability(
            n_type=capabilities.NodeCapabilityType.TCPSERVER, port=123
        )
        network_addr = address.NetworkAddress(
            address="127.0.0.1:0", timestamp=0, capabilities=[capa]
        )
        addr_payload = address.AddrPayload([network_addr])

        deserialized_addr_payload = address.AddrPayload.deserialize_from_bytes(
            addr_payload.to_array()
        )
        self.assertEqual(
            123, deserialized_addr_payload.addresses[0].capabilities[0].port
        )

    def test_equality(self):
        addr1 = address.NetworkAddress(address="127.0.0.1:0")
        addr2 = address.NetworkAddress(address="127.0.0.1:0")
        addr3 = address.NetworkAddress(address="127.0.0.2:0")
        self.assertTrue(addr1 == addr2)
        self.assertFalse(addr1 == addr3)
        self.assertFalse(addr1 == object())

    def test_various_dunder(self):
        addr1 = address.NetworkAddress(address="127.0.0.1:0")
        self.assertEqual("127.0.0.1:0", str(addr1))
        self.assertNotEqual("127.0.0.1", str(addr1))

        formatted = f"{addr1:>12}"
        self.assertEqual(" 127.0.0.1:0", formatted)

        # an object has to be hashable in order for it to be acceptable as a _dictionary key
        # this tests the __hash__ dunder
        try:
            x = {addr1: 123}
        except Exception as e:
            self.fail(f"Failed to add NetworkAddress to dictionary. Can't hash?: {e}")

        # test __repr__
        self.assertIn("127.0.0.1", repr(addr1))
        self.assertIn("(NEW)", repr(addr1))

    def test_properties(self):
        addr1 = address.NetworkAddress(address="127.0.0.1:30333")

        self.assertEqual("127.0.0.1", addr1.ip)
        self.assertEqual(30333, addr1.port)

        self.assertTrue(addr1.is_state_new)
        addr1.set_state_connected()
        self.assertFalse(addr1.is_state_new)

        self.assertTrue(addr1.is_state_connected)
        addr1.set_state_poor()
        self.assertFalse(addr1.is_state_connected)

        self.assertTrue(addr1.is_state_poor)
        addr1.set_state_dead()
        self.assertFalse(addr1.is_state_poor)

        self.assertTrue(addr1.is_state_dead)
        addr1.set_state_new()
        self.assertFalse(addr1.is_state_dead)


class BlockTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Transaction tx = new Transaction();
        tx.Nonce = 123;
        tx.SystemFee = 456;
        tx.NetworkFee = 789;
        tx.ValidUntilBlock = 1;
        tx.Attributes = new TransactionAttribute[0];
        tx.Signers = new Signer[] { new Signer() { Account = UInt160.Parse("0xe239c7228fa6b46cc0cf43623b2f934301d0b4f7"), Scopes = WitnessScope.CalledByEntry}};
        tx.Script = new byte[] { 0x1 };
        tx.Witnesses = new Witness[] {new Witness { InvocationScript = new byte[0], VerificationScript = new byte[] { 0x55 } }};

        Block b = new Block
        {
            Header = new Header
            {
                Version = 0,
                PrevHash = UInt256.Parse("0xf782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                Timestamp = 123,
                Nonce = 0,
                Index = 1,
                PrimaryIndex = 0,
                NextConsensus = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654"),
                Witness = new Witness { InvocationScript = new byte[0], VerificationScript = new byte[] { 0x55 } },
                MerkleRoot = UInt256.Zero
            },
            Transactions = new Transaction[] { tx }
        };
        b.Header.MerkleRoot = MerkleTree.ComputeRoot(b.Transactions.Select(p => p.Hash).ToArray());
        Console.WriteLine($"{b.Size}");
        Console.WriteLine($"{BitConverter.ToString(b.ToArray()).Replace("-", "")}");

        var trimmedBlock = new TrimmedBlock
        {
            Header = b.Header,
            Hashes = b.Transactions.Select(p => p.Hash).ToArray()
        };
        Console.WriteLine($"{trimmedBlock.Size}");
        Console.WriteLine($"{BitConverter.ToString(trimmedBlock.ToArray()).Replace("-", "")}");

        """
        cls.tx = transaction.Transaction(
            version=0,
            nonce=123,
            system_fee=456,
            network_fee=789,
            valid_until_block=1,
            attributes=[],
            signers=[
                verification.Signer(
                    types.UInt160.from_string(
                        "e239c7228fa6b46cc0cf43623b2f934301d0b4f7"
                    )
                )
            ],
            script=b"\x01",
            witnesses=[
                verification.Witness(invocation_script=b"", verification_script=b"\x55")
            ],
        )

        cls.header = block.Header(
            version=0,
            prev_hash=types.UInt256.from_string(
                "f782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"
            ),
            timestamp=123,
            nonce=0,
            index=1,
            primary_index=0,
            next_consensus=types.UInt160.from_string(
                "d7678dd97c000be3f33e9362e673101bac4ca654"
            ),
            witness=verification.Witness(
                invocation_script=b"", verification_script=b"\x55"
            ),
        )
        cls.block = block.Block(cls.header, transactions=[cls.tx])
        cls.block.rebuild_merkle_root()

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 168
        self.assertEqual(expected_len, len(self.block))

    def test_equals(self):
        self.assertFalse(None == self.block)
        self.assertFalse(self.block == object())

        # test different hashes
        modified_block = deepcopy(self.block)
        modified_block.header.timestamp = 1
        self.assertFalse(self.block == modified_block)
        self.assertTrue(self.block == self.block)

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            "000000000AAEB9C2C35A97AF7C054553CE64DA5E52FB530D6CB929E6AFF6EEB2FBC782F7120207967449B8E964E2E731A5CBF440CA39FABDD1EA47845B701790E301246B7B000000000000000000000000000000010000000054A64CAC1B1073E662933EF3E30B007CD98D67D70100015501007B000000C80100000000000015030000000000000100000001F7B4D00143932F3B6243CFC06CB4A68F22C739E20100010101000155"
        )
        self.assertEqual(expected_data, self.block.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_block = block.Block.deserialize_from_bytes(self.block.to_array())
        self.assertEqual(self.block.version, deserialized_block.version)
        self.assertEqual(self.block.prev_hash, deserialized_block.prev_hash)
        self.assertEqual(self.block.timestamp, deserialized_block.timestamp)
        self.assertEqual(self.block.nonce, deserialized_block.nonce)
        self.assertEqual(self.block.index, deserialized_block.index)
        self.assertEqual(self.block.primary_index, deserialized_block.primary_index)
        self.assertEqual(self.block.next_consensus, deserialized_block.next_consensus)
        self.assertEqual(
            self.block.witness.invocation_script,
            deserialized_block.witness.invocation_script,
        )
        self.assertEqual(
            self.block.witness.verification_script,
            deserialized_block.witness.verification_script,
        )
        self.assertEqual(1, len(deserialized_block.transactions))

    def test_deserialization_no_duplicate_transactions(self):
        # A block should not have duplicate transactions
        block_copy = deepcopy(self.block)
        block_copy.transactions.append(block_copy.transactions[0])
        with self.assertRaises(ValueError) as context:
            block.Block.deserialize_from_bytes(block_copy.to_array())
        self.assertIn(
            "Deserialization error - block contains duplicate transaction",
            str(context.exception),
        )

    def test_deserialization_wrong_merkle_root(self):
        block_copy = deepcopy(self.block)
        block_copy.header.merkle_root = types.UInt256.zero()
        with self.assertRaises(ValueError) as context:
            block.Block.deserialize_from_bytes(block_copy.to_array())
        self.assertIn(
            "Deserialization error - merkle root mismatch", str(context.exception)
        )

    def test_inventory_type(self):
        self.assertEqual(inventory.InventoryType.BLOCK, self.block.inventory_type)

    def test_trim(self):
        trimmed_block = self.block.trim()
        self.assertIsInstance(trimmed_block, block.TrimmedBlock)
        # captured from C#, see setUpClass() for the capture code
        expected_len = 146
        self.assertEqual(expected_len, len(trimmed_block))

        expected_data = binascii.unhexlify(
            "000000000AAEB9C2C35A97AF7C054553CE64DA5E52FB530D6CB929E6AFF6EEB2FBC782F7120207967449B8E964E2E731A5CBF440CA39FABDD1EA47845B701790E301246B7B000000000000000000000000000000010000000054A64CAC1B1073E662933EF3E30B007CD98D67D70100015501120207967449B8E964E2E731A5CBF440CA39FABDD1EA47845B701790E301246B"
        )
        self.assertEqual(expected_data, trimmed_block.to_array())

        deserialized_trimmed_block = block.TrimmedBlock.deserialize_from_bytes(
            trimmed_block.to_array()
        )
        self.assertEqual(trimmed_block.header, deserialized_trimmed_block.header)
        self.assertEqual(trimmed_block.hashes, deserialized_trimmed_block.hashes)
        self.assertEqual(1, len(deserialized_trimmed_block.hashes))


class SignerTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Signer co = new Signer();
        co.Account = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654");
        co.Scopes = WitnessScope.CustomContracts | WitnessScope.CustomGroups;
        co.AllowedContracts = new UInt160[] { UInt160.Parse("5b7074e873973a6ed3708862f219a6fbf4d1c411") };

        ECPoint p = ECPoint.Parse("026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc", ECCurve.Secp256r1);
        co.AllowedGroups = new ECPoint[] { p };
        co.Rules = new WitnessRule[0];

        Console.WriteLine($"{co.Size}");
        Console.WriteLine($"{BitConverter.ToString(co.ToArray()).Replace("-","")}");
        Console.Write(co.ToJson());
        """
        cls.signer = verification.Signer(
            types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654")
        )
        cls.signer.scope = (
            verification.WitnessScope.CUSTOM_CONTRACTS
            | verification.WitnessScope.CUSTOM_GROUPS
        )
        cls.signer.allowed_contracts = [
            types.UInt160.from_string("5b7074e873973a6ed3708862f219a6fbf4d1c411")
        ]
        point = crypto.ECPoint(
            binascii.unhexlify(
                "026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc"
            ),
            crypto.ECCCurve.SECP256R1,
            True,
        )
        cls.signer.allowed_groups = [point]

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 76
        self.assertEqual(expected_len, len(self.signer))

    def test_equals(self):
        self.assertFalse(self.signer == None)
        self.assertFalse(self.signer == object())
        signer2 = verification.Signer._serializable_init()
        self.assertFalse(self.signer == signer2)
        self.assertTrue(self.signer == self.signer)

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"54A64CAC1B1073E662933EF3E30B007CD98D67D7300111C4D1F4FBA619F2628870D36E3A9773E874705B01026241E7E26B38BB7154B8AD49458B97FB1C4797443DC921C5CA5774F511A2BBFC"
        )
        self.assertEqual(expected_data, self.signer.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_signer = verification.Signer.deserialize_from_bytes(
            self.signer.to_array()
        )
        self.assertEqual(self.signer.account, deserialized_signer.account)
        self.assertEqual(self.signer.scope, deserialized_signer.scope)
        self.assertEqual(
            self.signer.allowed_contracts, deserialized_signer.allowed_contracts
        )
        self.assertEqual(self.signer.allowed_groups, deserialized_signer.allowed_groups)

    def test_deserialization_invalid_scope(self):
        data = bytearray(self.signer.to_array())
        # the scope is serialized after the UInt160 `account` attribute
        # we modify that byte to an invalid scope type
        data[20] = 0xFF

        with self.assertRaises(ValueError) as context:
            verification.Signer.deserialize_from_bytes(data)
        self.assertEqual(
            "Deserialization error - invalid scope. GLOBAL scope not allowed with other scope types",
            str(context.exception),
        )

    def test_to_json(self):
        # captured from C#, see setUpClass() for the capture code
        expected = {
            "account": "0xd7678dd97c000be3f33e9362e673101bac4ca654",
            "scopes": "CustomContracts, CustomGroups",
            "allowedcontracts": ["0x5b7074e873973a6ed3708862f219a6fbf4d1c411"],
            "allowedgroups": [
                "026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc"
            ],
        }
        self.assertEqual(expected, self.signer.to_json())

    def test_from_json(self):
        captured = {
            "account": "0xd7678dd97c000be3f33e9362e673101bac4ca654",
            "scopes": "CustomContracts, CustomGroups",
            "allowedcontracts": ["0x5b7074e873973a6ed3708862f219a6fbf4d1c411"],
            "allowedgroups": [
                "026241e7e26b38bb7154b8ad49458b97fb1c4797443dc921c5ca5774f511a2bbfc"
            ],
        }
        signer = verification.Signer.from_json(captured)
        # relies on test_to_json() to pass
        self.assertEqual(captured, signer.to_json())


class FilterAddTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        FilterAddPayload fa = new FilterAddPayload { Data = new byte[] { 0x1, 0x2 } };
        Console.WriteLine($"{fa.Size}");
        Console.WriteLine($"{BitConverter.ToString(fa.ToArray()).Replace("-", "")}");
        """
        cls.filter = filterpayload.FilterAddPayload(b"\x01\x02")

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 3
        self.assertEqual(expected_len, len(self.filter))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b"020102")
        self.assertEqual(expected_data, self.filter.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_filter = filterpayload.FilterAddPayload.deserialize_from_bytes(
            self.filter.to_array()
        )
        self.assertEqual(self.filter.data, deserialized_filter.data)


class FilterLoadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        BloomFilter bf = new BloomFilter(8, 2, 345, new byte[] { 0x1, 0x2 });
        FilterLoadPayload fl = FilterLoadPayload.Create(bf);
        Console.WriteLine($"{fl.Size}");
        Console.WriteLine($"{BitConverter.ToString(fl.ToArray()).Replace("-", "")}");
        """
        bloom = crypto.BloomFilter(8, 2, 345, b"\x01\x02")
        cls.filter = filterpayload.FilterLoadPayload(bloom)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 7
        self.assertEqual(expected_len, len(self.filter))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b"01010259010000")
        self.assertEqual(expected_data, self.filter.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_filter = filterpayload.FilterLoadPayload.deserialize_from_bytes(
            self.filter.to_array()
        )
        self.assertEqual(self.filter.filter, deserialized_filter.filter)
        self.assertEqual(self.filter.K, deserialized_filter.K)
        self.assertEqual(self.filter.tweak, deserialized_filter.tweak)

    def test_deserialization_with_invalid_K_value(self):
        invalid_K = 51
        bloom = crypto.BloomFilter(8, invalid_K, 345, b"\x01\x02")
        filter = filterpayload.FilterLoadPayload(bloom)

        with self.assertRaises(ValueError) as context:
            filterpayload.FilterLoadPayload.deserialize_from_bytes(filter.to_array())
        self.assertIn(
            "Deserialization error - K exceeds limit of 50", str(context.exception)
        )


class GetBlocksPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        UInt256 hash_start = UInt256.Parse("0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01");
        GetBlocksPayload payload = GetBlocksPayload.Create(hash_start, 2);
        Console.WriteLine($"len: {payload.Size}");
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.hash_start = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"
        )
        cls.payload = block.GetBlocksPayload(hash_start=cls.hash_start, count=2)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 34
        self.assertEqual(expected_len, len(self.payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"01FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A40200"
        )
        self.assertEqual(expected_data, self.payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_payload = block.GetBlocksPayload.deserialize_from_bytes(
            self.payload.to_array()
        )
        self.assertEqual(self.payload.hash_start, deserialized_payload.hash_start)
        self.assertEqual(2, deserialized_payload.count)


class GetBlockByIndexPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        GetBlockByIndexPayload payload = GetBlockByIndexPayload.Create(1, 2);
        Console.WriteLine($"len: {payload.Size}");
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.payload = block.GetBlockByIndexPayload(index_start=1, count=2)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 6
        self.assertEqual(expected_len, len(self.payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b"010000000200")
        self.assertEqual(expected_data, self.payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_payload = block.GetBlockByIndexPayload.deserialize_from_bytes(
            self.payload.to_array()
        )
        self.assertEqual(self.payload.index_start, deserialized_payload.index_start)
        self.assertEqual(2, deserialized_payload.count)

    def test_deserialization_error(self):
        # test exceed max count
        payload = block.GetBlockByIndexPayload(
            index_start=1, count=block.HeadersPayload.MAX_HEADERS_COUNT + 1
        )
        with self.assertRaises(ValueError) as context:
            block.GetBlockByIndexPayload.deserialize_from_bytes(payload.to_array())
        self.assertIn("Deserialization error - invalid count", str(context.exception))

        # test 0 count
        payload = block.GetBlockByIndexPayload(index_start=1, count=0)
        with self.assertRaises(ValueError) as context:
            block.GetBlockByIndexPayload.deserialize_from_bytes(payload.to_array())
        self.assertIn("Deserialization error - invalid count", str(context.exception))

        # test negative count
        payload = block.GetBlockByIndexPayload(index_start=1, count=-10)
        with self.assertRaises(ValueError) as context:
            block.GetBlockByIndexPayload.deserialize_from_bytes(payload.to_array())
        self.assertIn("Deserialization error - invalid count", str(context.exception))


class HeaderTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Header h = new Header
        {
            Version = 0,
            PrevHash = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"),
            MerkleRoot = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"),
            Timestamp = 0,
            Nonce = 0,
            Index = 123,
            PrimaryIndex = 0,
            NextConsensus = UInt160.Parse("0xe239c7228fa6b46cc0cf43623b2f934301d0b4f7"),
            Witness = new Witness
            {
                InvocationScript = new byte[] { 0x1, 0x2},
                VerificationScript = new byte[] { 0x3, 0x4}
            }
        };
        Console.WriteLine(h.Hash);
        Console.WriteLine(h.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(h.ToArray()).Replace("-", "")}\'");
        """
        version = 0
        previous_hash = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"
        )
        merkleroot = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"
        )
        timestamp = 0
        nonce = 0
        index = 123
        primary_index = 0
        next_consensus = types.UInt160.from_string(
            "e239c7228fa6b46cc0cf43623b2f934301d0b4f7"
        )
        witness = verification.Witness(
            invocation_script=b"\x01\x02", verification_script=b"\x03\x04"
        )

        cls.header = block.Header(
            version,
            previous_hash,
            timestamp,
            nonce,
            index,
            primary_index,
            next_consensus,
            witness,
            merkleroot,
        )

    def test_len_and_hash(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 116
        expected_hash = types.UInt256.from_string(
            "f8e49a63506ed1273ec79ec6fc0134803aafc251c7e67ffaf355a443617672c5"
        )
        self.assertEqual(expected_len, len(self.header))
        self.assertEqual(expected_hash, self.header.hash())

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"0000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B00000000F7B4D00143932F3B6243CFC06CB4A68F22C739E201020102020304"
        )
        self.assertEqual(expected_data, self.header.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_header = block.Header.deserialize_from_bytes(
            self.header.to_array()
        )
        self.assertEqual(self.header.hash(), deserialized_header.hash())
        self.assertEqual(self.header.version, deserialized_header.version)
        self.assertEqual(self.header.prev_hash, deserialized_header.prev_hash)
        self.assertEqual(self.header.merkle_root, deserialized_header.merkle_root)
        self.assertEqual(self.header.timestamp, deserialized_header.timestamp)
        self.assertEqual(self.header.nonce, deserialized_header.nonce)
        self.assertEqual(self.header.index, deserialized_header.index)
        self.assertEqual(self.header.next_consensus, deserialized_header.next_consensus)
        self.assertEqual(
            self.header.witness.invocation_script,
            deserialized_header.witness.invocation_script,
        )
        self.assertEqual(
            self.header.witness.verification_script,
            deserialized_header.witness.verification_script,
        )

    def test_deserialization_failure1(self):
        # there should be a 1 byte witness object count (fixed to value 1) before the actual witness object.
        # see https://github.com/neo-project/neo/issues/1128
        raw_data = binascii.unhexlify(
            b"0000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B00000000F7B4D00143932F3B6243CFC06CB4A68F22C739E200020102020304"
        )
        deserialized_header = block.Header._serializable_init()

        with self.assertRaises(ValueError) as context:
            with serialization.BinaryReader(raw_data) as br:
                deserialized_header.deserialize(br)
        self.assertIn("Deserialization error", str(context.exception))
        self.assertIn("Witness object count is 0 must be 1", str(context.exception))

    def test_equals(self):
        self.assertFalse(None == self.header)
        self.assertFalse(self.header == object())

        # test different hashes
        modified_header = deepcopy(self.header)
        modified_header.timestamp = 1
        self.assertFalse(self.header == modified_header)
        self.assertTrue(self.header == self.header)


class HeadersPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Header h1 = new Header
        {
            Version = 0,
            PrevHash = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"),
            MerkleRoot = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"),
            Timestamp = 0,
            Nonce = 0,
            Index = 123,
            PrimaryIndex = 0,
            NextConsensus = UInt160.Parse("8a2b438eaca8b4b2ab6b4524b5a69a45d920c351"),
            Witness = new Witness
            {
                InvocationScript = new byte[] {0x1, 0x2},
                VerificationScript = new byte[] {0x3, 0x4}
            }
        };
        Header h2;
        using (BinaryReader reader = new BinaryReader(new MemoryStream(h1.ToArray())))
        {
            h2 = reader.ReadSerializable<Header>();
        }

        HeadersPayload hp = HeadersPayload.Create(new Header[] { h1, h2 });
        Console.WriteLine(hp.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(hp.ToArray()).Replace("-", "")}\'");
        """
        version = 0
        previous_hash = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"
        )
        merkleroot = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"
        )
        timestamp = 0
        nonce = 0
        index = 123
        primary_index = 0
        next_consensus = types.UInt160.from_string(
            "8a2b438eaca8b4b2ab6b4524b5a69a45d920c351"
        )
        witness = verification.Witness(
            invocation_script=b"\x01\x02", verification_script=b"\x03\x04"
        )

        h1 = block.Header(
            version,
            previous_hash,
            timestamp,
            nonce,
            index,
            primary_index,
            next_consensus,
            witness,
            merkleroot,
        )
        h2 = block.Header(
            version,
            previous_hash,
            timestamp,
            nonce,
            index,
            primary_index,
            next_consensus,
            witness,
            merkleroot,
        )
        cls.payload = block.HeadersPayload([h1, h2])

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 233
        self.assertEqual(expected_len, len(self.payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"020000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B0000000051C320D9459AA6B524456BABB2B4A8AC8E432B8A010201020203040000000001FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4000000000000000000000000000000007B0000000051C320D9459AA6B524456BABB2B4A8AC8E432B8A01020102020304"
        )
        self.assertEqual(expected_data, self.payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_payload = block.HeadersPayload.deserialize_from_bytes(
            self.payload.to_array()
        )
        self.assertEqual(len(self.payload.headers), len(deserialized_payload.headers))
        self.assertEqual(2, len(self.payload.headers))
        self.assertIsInstance(deserialized_payload.headers[0], block.Header)
        self.assertIsInstance(deserialized_payload.headers[1], block.Header)


class InventoryPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var u1 = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01");
        var u2 = UInt256.Parse("a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02");
        var payload = InvPayload.Create(InventoryType.Block, new UInt256[] { u1, u2 });
        Console.WriteLine(payload.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.u1 = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff01"
        )
        cls.u2 = types.UInt256.from_string(
            "a400ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff00ff02"
        )
        cls.inv = inventory.InventoryPayload(
            inventory.InventoryType.BLOCK, [cls.u1, cls.u2]
        )

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 66
        self.assertEqual(expected_len, len(self.inv))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"2C0201FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A402FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00A4"
        )
        self.assertEqual(expected_data, self.inv.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_inv_payload = inventory.InventoryPayload.deserialize_from_bytes(
            self.inv.to_array()
        )
        self.assertEqual(self.inv.type, deserialized_inv_payload.type)
        self.assertEqual(len(self.inv.hashes), len(deserialized_inv_payload.hashes))
        self.assertEqual(self.inv.hashes[0], deserialized_inv_payload.hashes[0])


class MerkleBlockPayloadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Transaction tx = new Transaction();
        tx.Nonce = 123;
        tx.SystemFee = 456;
        tx.NetworkFee = 789;
        tx.ValidUntilBlock = 1;
        tx.Attributes = new TransactionAttribute[0];
        tx.Signers = new Signer[] { new Signer() { Account = UInt160.Parse("0xe239c7228fa6b46cc0cf43623b2f934301d0b4f7"), Scopes = WitnessScope.CalledByEntry}};
        tx.Script = new byte[] { 0x1 };
        tx.Witnesses = new Witness[0];

        Block b = new Block
        {
            Header = new Header
            {
                Version = 0,
                PrevHash = UInt256.Parse("0xf782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"),
                Timestamp = 123,
                Nonce = 0,
                Index = 1,
                PrimaryIndex = 0,
                NextConsensus = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654"),
                Witness = new Witness { InvocationScript = new byte[0], VerificationScript = new byte[] { 0x55 } },
                MerkleRoot = UInt256.Zero
            },
            Transactions = new Transaction[] { tx }
        };
        b.Header.MerkleRoot = MerkleTree.ComputeRoot(b.Transactions.Select(p => p.Hash).ToArray());

        byte[] bytes = { 0x1 };
        BitArray flags = new BitArray(bytes);
        MerkleBlockPayload mbp = MerkleBlockPayload.Create(b, flags);
        Console.WriteLine($"b\'{BitConverter.ToString(mbp.ToArray()).Replace("-", "")}\'");
        """

        cls.tx = transaction.Transaction(
            version=0,
            nonce=123,
            system_fee=456,
            network_fee=789,
            valid_until_block=1,
            attributes=[],
            signers=[
                verification.Signer(
                    types.UInt160.from_string(
                        "e239c7228fa6b46cc0cf43623b2f934301d0b4f7"
                    )
                )
            ],
            script=b"\x01",
            witnesses=[
                verification.Witness(invocation_script=b"", verification_script=b"\x55")
            ],
        )

        cls.header = block.Header(
            version=0,
            prev_hash=types.UInt256.from_string(
                "f782c7fbb2eef6afe629b96c0d53fb525eda64ce5345057caf975ac3c2b9ae0a"
            ),
            timestamp=123,
            nonce=0,
            index=1,
            primary_index=0,
            next_consensus=types.UInt160.from_string(
                "d7678dd97c000be3f33e9362e673101bac4ca654"
            ),
            witness=verification.Witness(
                invocation_script=b"", verification_script=b"\x55"
            ),
        )
        cls.block = block.Block(cls.header, transactions=[cls.tx])
        cls.block.rebuild_merkle_root()
        flags = bitarray()
        flags.frombytes(b"\x01")
        cls.merkle_payload = block.MerkleBlockPayload(cls.block, flags)

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 152
        self.assertEqual(expected_len, len(self.merkle_payload))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"000000000AAEB9C2C35A97AF7C054553CE64DA5E52FB530D6CB929E6AFF6EEB2FBC782F7120207967449B8E964E2E731A5CBF440CA39FABDD1EA47845B701790E301246B7B000000000000000000000000000000010000000054A64CAC1B1073E662933EF3E30B007CD98D67D7010001550101120207967449B8E964E2E731A5CBF440CA39FABDD1EA47845B701790E301246B0101"
        )
        self.assertEqual(expected_data, self.merkle_payload.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_merkle_payload = block.MerkleBlockPayload.deserialize_from_bytes(
            self.merkle_payload.to_array()
        )
        # not testing all properties again. It re-uses the same block as created in the Block test case
        # only testing new properties
        self.assertEqual(
            self.merkle_payload.tx_count, deserialized_merkle_payload.tx_count
        )
        self.assertEqual(
            len(self.merkle_payload.hashes), len(deserialized_merkle_payload.hashes)
        )
        self.assertEqual(
            self.merkle_payload.hashes[0], deserialized_merkle_payload.hashes[0]
        )
        self.assertEqual(self.merkle_payload.flags, deserialized_merkle_payload.flags)


class PingTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var payload = PingPayload.Create(123, 456);
        payload.Timestamp = 888;
        Console.WriteLine(payload.Size);
        Console.WriteLine($"b\'{BitConverter.ToString(payload.ToArray()).Replace("-", "")}\'");
        """
        cls.ping = ping.PingPayload(123)
        cls.ping.nonce = 456
        cls.ping.timestamp = 888

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 12
        self.assertEqual(expected_len, len(self.ping))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b"7B00000078030000C8010000")
        self.assertEqual(expected_data, self.ping.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_ping = ping.PingPayload.deserialize_from_bytes(
            self.ping.to_array()
        )
        self.assertEqual(self.ping.timestamp, deserialized_ping.timestamp)
        self.assertEqual(self.ping.nonce, deserialized_ping.nonce)
        self.assertEqual(self.ping.current_height, deserialized_ping.current_height)


class TestAttributeTypeEnum(enum.IntEnum):
    TYPE_ZERO = 0x00


class TestTXAttribute(transaction.TransactionAttribute):
    def __init__(self):
        super(TestTXAttribute, self).__init__()

        self.type_ = TestAttributeTypeEnum.TYPE_ZERO
        self.test_member = True

    def __len__(self):
        size_bool = 1
        return super(TestTXAttribute, self).__len__() + size_bool

    def _deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        self.test_member = reader.read_bool()

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_bool(self.test_member)


class TransactionAttributeTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Requires a custom attribute

        class TestTXAttribute : TransactionAttribute
        {
            public override TransactionAttributeType Type { get; }
            public override bool AllowMultiple { get => true; }

            public bool test_member = true;

            public override int Size { get => base.Size + sizeof(bool); }

            protected override void DeserializeWithoutType(BinaryReader reader)
            {
                test_member = reader.ReadBoolean();
            }

            protected override void SerializeWithoutType(BinaryWriter writer)
            {
                writer.Write(test_member);
            }
        }

        TransactionAttribute ta = new TestTXAttribute();
        Console.WriteLine($"{ta.Size}");
        Console.WriteLine($"{BitConverter.ToString(ta.ToArray()).Replace("-", "")}");
        """
        cls.attribute = TestTXAttribute()

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 2
        self.assertEqual(expected_len, len(self.attribute))

    def test_equals(self):
        self.assertFalse(self.attribute == None)
        self.assertFalse(self.attribute == object())
        self.assertTrue(self.attribute == TestTXAttribute())

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b"0001")
        self.assertEqual(expected_data, self.attribute.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_attribute = self.attribute.deserialize_from_bytes(
            self.attribute.to_array()
        )
        self.assertEqual(self.attribute.type_, deserialized_attribute.type_)
        self.assertEqual(self.attribute.test_member, deserialized_attribute.test_member)

    def test_deserialization_wrong_type(self):
        stream_with_type_one = b"\x01"
        attribute = TestTXAttribute()
        with self.assertRaises(ValueError) as context:
            attribute.deserialize_from_bytes(stream_with_type_one)
        self.assertEqual(
            "Deserialization error - transaction attribute type mismatch",
            str(context.exception),
        )

    def test_deserialization_from(self):
        stream_with_type_zero = b"\x00\x01"
        with serialization.BinaryReader(stream_with_type_zero) as reader:
            ta = transaction.TransactionAttribute.deserialize_from(reader)
        self.assertIsInstance(ta, TestTXAttribute)

    def test_deserialization_from_failure(self):
        stream_with_invalid_type = b"\xFF"
        with serialization.BinaryReader(stream_with_invalid_type) as reader:
            with self.assertRaises(ValueError) as context:
                transaction.TransactionAttribute.deserialize_from(reader)
            self.assertEqual(
                "Deserialization error - unknown transaction attribute type",
                str(context.exception),
            )


class TransactionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Transaction tx = new Transaction();
        tx.Nonce = 123;
        tx.SystemFee = 456;
        tx.NetworkFee = 789;
        tx.ValidUntilBlock = 1;
        tx.Attributes = new TransactionAttribute[] { };

        Signer co = new Signer();
        co.Account = UInt160.Parse("0xd7678dd97c000be3f33e9362e673101bac4ca654");
        co.Scopes = WitnessScope.None;
        tx.Signers = new Signer[] { co };

        tx.Script = new byte[] { 0x1, 0x2 };
        tx.Witnesses = new Witness[] { new Witness { InvocationScript=new byte[0], VerificationScript = new byte[] { 0x55 } } };

        Console.WriteLine($"{tx.Size}");
        Console.WriteLine($"{BitConverter.ToString(tx.ToArray()).Replace("-", "")}");
        Console.WriteLine(tx.Hash);
        Console.WriteLine(tx.FeePerByte);
        """
        signer = verification.Signer(
            account=types.UInt160.from_string(
                "d7678dd97c000be3f33e9362e673101bac4ca654"
            ),
            scope=verification.WitnessScope.NONE,
        )

        witness = verification.Witness(
            invocation_script=b"", verification_script=b"\x55"
        )

        cls.tx = transaction.Transaction(
            version=0,
            nonce=123,
            system_fee=456,
            network_fee=789,
            valid_until_block=1,
            attributes=[],
            signers=[signer],
            script=b"\x01\x02",
            witnesses=[witness],
        )

    def tearDown(cls) -> None:
        settings.reset_settings_to_default()

    def test_len_and_hash(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 55
        expected_hash = types.UInt256.from_string(
            "da0343daadf88f95ece657fed6c20e05256d37f0d68757034da1d34f534d2c2c"
        )
        self.assertEqual(expected_len, len(self.tx))
        self.assertEqual(expected_hash, self.tx.hash())

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"007B000000C8010000000000001503000000000000010000000154A64CAC1B1073E662933EF3E30B007CD98D67D7000002010201000155"
        )
        self.assertEqual(expected_data, self.tx.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_tx = transaction.Transaction.deserialize_from_bytes(
            self.tx.to_array()
        )
        self.assertEqual(self.tx.version, deserialized_tx.version)
        self.assertEqual(self.tx.nonce, deserialized_tx.nonce)
        self.assertEqual(self.tx.system_fee, deserialized_tx.system_fee)
        self.assertEqual(self.tx.network_fee, deserialized_tx.network_fee)
        self.assertEqual(self.tx.valid_until_block, deserialized_tx.valid_until_block)
        self.assertEqual(len(self.tx.attributes), len(deserialized_tx.attributes))
        self.assertEqual(len(self.tx.signers), len(deserialized_tx.signers))
        self.assertEqual(self.tx.signers[0].account, deserialized_tx.signers[0].account)
        self.assertEqual(self.tx.signers[0].scope, deserialized_tx.signers[0].scope)
        self.assertEqual(self.tx.script, deserialized_tx.script)
        self.assertEqual(len(self.tx.witnesses), len(deserialized_tx.witnesses))
        self.assertEqual(
            self.tx.witnesses[0].invocation_script,
            deserialized_tx.witnesses[0].invocation_script,
        )
        self.assertEqual(
            self.tx.witnesses[0].verification_script,
            deserialized_tx.witnesses[0].verification_script,
        )

    def test_deserialization_version_error(self):
        tx = deepcopy(self.tx)
        tx.version = 1
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - invalid version", str(context.exception)
        )

    def test_deserialization_system_fee_error(self):
        tx = deepcopy(self.tx)
        tx.system_fee = -1
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - negative system fee", str(context.exception)
        )

    def test_deserialization_network_fee_error(self):
        tx = deepcopy(self.tx)
        tx.network_fee = -1
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - negative network fee", str(context.exception)
        )

    def test_deserialization_script_length_error(self):
        tx = deepcopy(self.tx)
        tx.script = b""
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - invalid script length 0", str(context.exception)
        )

    def test_deserialization_duplicate_attribute_error(self):
        tx = deepcopy(self.tx)
        tx.attributes = [TestTXAttribute(), TestTXAttribute()]
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - duplicate transaction attribute",
            str(context.exception),
        )

    def test_deserialization_empty_signers_error(self):
        tx = deepcopy(self.tx)
        tx.signers = []
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - signers can't be empty", str(context.exception)
        )

    def test_deserialization_duplicate_signer(self):
        tx = deepcopy(self.tx)
        tx.signers[0].scope = verification.WitnessScope.GLOBAL
        tx.signers.append(deepcopy(tx.signers[0]))
        with self.assertRaises(ValueError) as context:
            transaction.Transaction.deserialize_from_bytes(tx.to_array())
        self.assertEqual(
            "Deserialization error - duplicate signer", str(context.exception)
        )

    def test_fee_per_byte(self):
        # captured from C#, see setUpClass() for the capture code
        expected_fee = 14
        self.assertEqual(expected_fee, self.tx.fee_per_byte())

    def test_equals(self):
        self.assertFalse(None == self.tx)
        self.assertFalse(self.tx == object())

        # test different hashes
        modified_tx = deepcopy(self.tx)
        modified_tx.version = 1
        self.assertFalse(self.tx == modified_tx)
        self.assertTrue(self.tx == self.tx)

    def test_inventory_type(self):
        self.assertEqual(inventory.InventoryType.TX, self.tx.inventory_type)

    def test_special_serialization(self):
        tx_special = deepcopy(self.tx)
        tx_special.block_height = 1
        with serialization.BinaryWriter() as bw:
            tx_special.serialize_special(bw)
            tx_special_bytes = bw.to_array()

        with serialization.BinaryReader(tx_special_bytes) as br:
            tx_from_bytes = transaction.Transaction._serializable_init()
            tx_from_bytes.deserialize_special(br)
        self.assertEqual(tx_special.block_height, tx_from_bytes.block_height)

    def test_sender(self):
        self.assertEqual(self.tx.signers[0].account, self.tx.sender)

        tx = deepcopy(self.tx)
        tx.signers = []
        self.assertEqual(types.UInt160.zero(), tx.sender)

    def test_protocol_magic(self):
        # test proper initialization
        tx = transaction.Transaction._serializable_init()
        # test default
        self.assertEqual(0x4F454E, tx.protocol_magic)
        # test init supplied
        tx = transaction.Transaction(0, 0, 0, 0, 0, protocol_magic=123)
        self.assertEqual(123, tx.protocol_magic)
        # test settings supplied
        settings.network.magic = 456
        tx = transaction.Transaction._serializable_init()
        self.assertEqual(456, tx.protocol_magic)

    def test_scripthashes_for_verifying(self):
        tx = transaction.Transaction._serializable_init()

        account1 = types.UInt160.from_string("d7678dd97c000be3f33e9362e673101bac4ca654")
        account2 = types.UInt160.zero()
        signer1 = verification.Signer(
            account=account1, scope=verification.WitnessScope.NONE
        )
        signer2 = verification.Signer(
            account=account2, scope=verification.WitnessScope.NONE
        )

        tx.signers = [signer1, signer2]

        hashes = tx.get_script_hashes_for_verifying(None)
        self.assertEqual([account1, account2], hashes)

    def test_to_json(self):
        tx = deepcopy(self.tx)
        tx.attributes = [transaction.HighPriorityAttribute()]

        expected = {
            "hash": "0x61a72c37793ea32b9181105931e7a935eb57c915e9dbeb84dc37068c21df2b20",
            "size": 56,
            "version": 0,
            "nonce": 123,
            "sender": "NTdZAZXWiMSDUTkEFU3nHpT7FEEwwtjwfC",
            "sysfee": "456",
            "netfee": "789",
            "validuntilblock": 1,
            "signers": [
                {
                    "account": "0xd7678dd97c000be3f33e9362e673101bac4ca654",
                    "scopes": "None",
                }
            ],
            "attributes": [{"type": "HighPriority"}],
            "script": "AQI=",
            "witnesses": [{"invocation": "", "verification": "VQ=="}],
        }
        self.assertEqual(expected, tx.to_json())

    def test_from_json(self):
        expected = expected = {
            "hash": "0x61a72c37793ea32b9181105931e7a935eb57c915e9dbeb84dc37068c21df2b20",
            "size": 56,
            "version": 0,
            "nonce": 123,
            "sender": "NTdZAZXWiMSDUTkEFU3nHpT7FEEwwtjwfC",
            "sysfee": "456",
            "netfee": "789",
            "validuntilblock": 1,
            "signers": [
                {
                    "account": "0xd7678dd97c000be3f33e9362e673101bac4ca654",
                    "scopes": "None",
                }
            ],
            "attributes": [{"type": "HighPriority"}],
            "script": "AQI=",
            "witnesses": [{"invocation": "", "verification": "VQ=="}],
        }
        tx = transaction.Transaction.from_json(expected)
        self.assertEqual(expected, tx.to_json())


class VersionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        var capabilities = new List<NodeCapability>
        {
            new FullNodeCapability(123)
        };

        capabilities.Add(new ServerCapability(NodeCapabilityType.TcpServer, (ushort)456));
        NodeCapability[] caps = capabilities.ToArray();
        ISerializable message = VersionPayload.Create(888, "my-user-agent", capabilities.ToArray());
        ((VersionPayload)message).Magic = 769;
        ((VersionPayload)message).Timestamp = 0;
        Console.WriteLine($"b\'{BitConverter.ToString(message.ToArray()).Replace("-", "")}\'");
        """
        settings.network.magic = 769
        capa = [
            capabilities.FullNodeCapability(start_height=123),
            capabilities.ServerCapability(
                capabilities.NodeCapabilityType.TCPSERVER, port=456
            ),
        ]
        cls.vp = version.VersionPayload(
            nonce=888, user_agent="my-user-agent", capabilities=capa
        )
        cls.vp.timestamp = 0

    @classmethod
    def tearDownClass(cls) -> None:
        settings.reset_settings_to_default()

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 39
        self.assertEqual(expected_len, len(self.vp))

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(
            b"010300000000000000000000780300000D6D792D757365722D6167656E7402107B00000001C801"
        )
        self.assertEqual(expected_data, self.vp.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_vp = version.VersionPayload.deserialize_from_bytes(
            self.vp.to_array()
        )
        self.assertEqual(len(self.vp.capabilities), len(deserialized_vp.capabilities))
        self.assertIsInstance(
            deserialized_vp.capabilities[0], capabilities.FullNodeCapability
        )
        self.assertEqual(
            self.vp.capabilities[0].start_height,
            deserialized_vp.capabilities[0].start_height,
        )
        self.assertIsInstance(
            deserialized_vp.capabilities[1], capabilities.ServerCapability
        )
        self.assertEqual(
            self.vp.capabilities[1].port, deserialized_vp.capabilities[1].port
        )


class WitnessTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Witness witness = new Witness();
        witness.VerificationScript = new byte[] { 0x01, 0x02 };
        witness.InvocationScript = new byte[] { 0x03, 0x04 };
        Console.WriteLine($"b\'{BitConverter.ToString(witness.ToArray()).Replace("-","")}\'");
        Console.Write(witness.ToJson());
        """
        cls.witness = verification.Witness(
            verification_script=b"\x01\x02", invocation_script=b"\x03\x04"
        )

    def test_len(self):
        # captured from C#, see setUpClass() for the capture code
        expected_len = 6
        self.assertEqual(expected_len, len(self.witness))

    def test_script_hash(self):
        # captured from C#, see setUpClass() for the capture code
        expected_script_hash = types.UInt160.from_string(
            "9132aaf67ab75c0a604419d920c5cb91e149cc15"
        )
        h = self.witness.script_hash()
        self.assertIsInstance(h, types.UInt160)
        self.assertEqual(expected_script_hash, h)

    def test_serialization(self):
        # captured from C#, see setUpClass() for the capture code
        expected_data = binascii.unhexlify(b"020304020102")
        self.assertEqual(expected_data, self.witness.to_array())

    def test_deserialization(self):
        # if the serialization() test for this class passes, we can use that as a reference to test deserialization against
        deserialized_witness = verification.Witness.deserialize_from_bytes(
            self.witness.to_array()
        )
        self.assertEqual(
            self.witness.verification_script, deserialized_witness.verification_script
        )
        self.assertEqual(
            self.witness.invocation_script, deserialized_witness.invocation_script
        )

    def test_to_json(self):
        # captured from C#, see setUpClass() for the capture code
        expected = {"invocation": "AwQ=", "verification": "AQI="}
        self.assertEqual(expected, self.witness.to_json())

        expected_empty = {"invocation": "", "verification": ""}
        empty_witness = verification.Witness(b"", b"")
        self.assertEqual(expected_empty, empty_witness.to_json())

    def test_from_json(self):
        witness_json = {"invocation": "AwQ=", "verification": "AQI="}
        witness = verification.Witness.from_json(witness_json)
        self.assertEqual(witness_json, witness.to_json())


class EmptyPayloadTestCase(unittest.TestCase):
    def test_empty(self):
        e = empty.EmptyPayload()
        self.assertEqual(0, len(e))
        self.assertEqual(0, len(e.to_array()))
        bogus_data = b"\x01\x02\x03"
        self.assertIsInstance(
            empty.EmptyPayload.deserialize_from_bytes(bogus_data), empty.EmptyPayload
        )



import unittest
from neo3.contracts.callflags import CallFlags


class CallFlagsTestCase(unittest.TestCase):
    def test_parsing_from_string(self):
        tests = [
            ("None", CallFlags.NONE),
            ("ReadStates", CallFlags.READ_STATES),
            ("WriteStates", CallFlags.WRITE_STATES),
            ("AllowCall", CallFlags.ALLOW_CALL),
            ("AllowNotify", CallFlags.ALLOW_NOTIFY),
            ("States", CallFlags.STATES),
            ("ReadOnly", CallFlags.READ_ONLY),
            ("All", CallFlags.ALL),
        ]
        for input, expected in tests:
            self.assertEqual(expected, CallFlags.from_csharp_name(input))

    def test_parsing_invalid_input(self):
        with self.assertRaises(ValueError) as context:
            CallFlags.from_csharp_name("bla")
        self.assertEqual(
            "bla is not a valid member of CallFlags", str(context.exception)
        )

    def test_parsing_from_string_with_multiple_flags(self):
        input = "ReadStates, AllowCall"
        cf = CallFlags.from_csharp_name(input)
        expected = CallFlags.READ_STATES | CallFlags.ALLOW_CALL
        self.assertNotEqual(CallFlags.READ_STATES, cf)
        self.assertNotEqual(CallFlags.ALLOW_CALL, cf)
        self.assertEqual(expected, cf)
